#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <unistd.h>
#include <sys/types.h>
#include <fcntl.h>
#include <pthread.h>
#include <string.h>

/* Prototype */
void *ParallelFunc(void *arg);

/*-------------------------------------
 * This utility creates a file, and
 * outputs a profile of disk write
 * throughput on a per-packet basis. 
 * Remove fsync() to see the huge
 * latency during large kernel buffer
 * flushes.
 *------------------------------------*/
main(int argc, char *argv[])
{
	/* Buffer and stream */
	FILE *fp;
	char *pBuffer;
	long i, dataSizeMB = 0, bufferSize = 0;
	char pFileName[100];
	int fd;

	/* Only used if user-buffered */
	//int bigBufSize = 600000000;
	//char *pBigBuffer = new char[bigBufSize]; 

	/* Timing and statistics */
	struct timeval start, end, beginWrite, endWrite;
	long thruArray[100000];
	int index;
	long duration;

	/* Threading */
	pthread_t parallelThread;
	bool stopThread = false;
	if(pthread_create(&parallelThread, NULL, ParallelFunc, (void *)&stopThread) != 0)
	{
		printf("\nError: cannot create thread!\n\n");
		fflush(stdout);
		exit(0);
	}

	if(argc != 3)
	{
		printf("\nUse format:\n\ncreate 100 1000 for 100MB file written in 1 MB chunks\n\n");
		exit(0);
	}

	dataSizeMB = atoi(argv[1]);
	bufferSize = atoi(argv[2]);
	
	sprintf(pFileName, "send%dM.dat", dataSizeMB);
	fp = fopen64(pFileName, "w");
	fd = fileno(fp);
	
	pBuffer = new char[bufferSize*1000];

	/* Buffer options */

	/* No buffering */
	//if(setvbuf(fp, NULL, _IONBF, 0) == 0)
		//printf("SUCCESS!!");

	/* User-buffered */
	//if(setvbuf(fp, pBigBuffer, _IOFBF, bigBufSize) == 0)
		//printf("SUCCESS!!");

	/* Kernel buffering, with specified size */
	//if(setvbuf(fp, NULL, _IOFBF, bigBufSize) == 0)
		//printf("SUCCESS!!");

	fflush(stdout);


	for(i = 0; i < bufferSize*1000; i ++)
		pBuffer[i] = 'A';	

	gettimeofday(&start, NULL);
	gettimeofday(&beginWrite,NULL);
	
	for(i = 0; i < dataSizeMB*1000/(bufferSize); i ++)
	{
		char *buff = new char[bufferSize*1000*10];
		
		for(int j = 0; j < 10; j++)
			memcpy(buff + (j*bufferSize*1000), pBuffer, bufferSize*1000);
		fwrite(pBuffer, 1, bufferSize*1000,fp);
		//if(i%4 == 0)
			//fsync(fd);
		delete [] buff;
		gettimeofday(&endWrite,NULL);
		thruArray[index++] = (endWrite.tv_sec-beginWrite.tv_sec)*1000000 + (endWrite.tv_usec-beginWrite.tv_usec);
		
		gettimeofday(&beginWrite,NULL);
	}

	fsync(fd);
	fclose(fp);
	//stopThread = true;

	gettimeofday(&endWrite,NULL);
	thruArray[index++] = (endWrite.tv_sec-beginWrite.tv_sec)*1000000 + (endWrite.tv_usec-beginWrite.tv_usec);

	/* End of execution statistics */
	gettimeofday(&end, NULL);
	duration = (end.tv_sec-start.tv_sec)*1000000 + (end.tv_usec-start.tv_usec);
	printf("\nWriting %ld MBytes in %lf seconds\n", dataSizeMB, duration/1000000.0);
	printf("\nThroughput for file writing: %lf Mbps\n\n", (double)dataSizeMB*1000000.0*8.0/(double)duration);

	/* Print times for each fwrite call */
	for(i = 0; i < index; i+=1)
	{
		if(thruArray[i] > 1000000)
			printf("FWRITE: %7ld ", thruArray[i]); 
	}
	pthread_join(parallelThread, NULL);
}

void *ParallelFunc(void *arg)
{
	bool *stopThread = (bool *)arg;
	struct timeval start, end, beginWrite, endWrite;
	long thruArray[500000];
	int index = 0;
	char *datagram = new char[1500];
	for(int i = 0; i < 1500; i++)
		datagram[i] = 'A';
	
		
	gettimeofday(&beginWrite,NULL);
	gettimeofday(&start,NULL);
	char **memLeak;
	memLeak = new char*[500000];
	for(int i = 0; i < 500000; i++)
	{
		memLeak[i] = new char[1500]; 
		//for(int j = 0; j < 1500; j++)
			//memLeak[i][j] = 'A';
		memcpy(memLeak[i],datagram,1500);
		//sleep(1);
		gettimeofday(&endWrite,NULL);
		thruArray[index++] = (endWrite.tv_sec-beginWrite.tv_sec)*1000000 + (endWrite.tv_usec-beginWrite.tv_usec);
		gettimeofday(&beginWrite,NULL);
	}
	gettimeofday(&end,NULL);

	long duration = (end.tv_sec-start.tv_sec)*1000000 + (end.tv_usec-start.tv_usec);
	printf("\nAllocating %ld MBytes in %lf seconds\n", 700, duration/1000000.0);
	printf("\nThroughput for allocating: %lf Mbps\n\n", (double)700.0*1000000.0*8.0/(double)duration);

	printf("\nHigh Latencies: \n"); 
	/*for(int i = 0; i < index; i+=1)
	{
		if(thruArray[i] > 15000)
			printf("%7ld ", thruArray[i]); 
	}*/
}


