//Developer:	Qishi Wu
//Date:			June 2004
//Place:		CSMD @ ORNL

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/time.h>
#include <fcntl.h>

#include "ioutil.h"

ssize_t myread(int fd, char *ptr)
{
	static int	read_cnt = 0;
	static char *read_ptr;
	static char read_buf[MAX_LINE_LEN];

	if(read_cnt <= 0)
	{
	again:
		if((read_cnt = read(fd, read_buf, sizeof(read_buf))) < 0)
		{
			if(errno == EINTR)	
				goto again;
			return(-1);
		}
		else if(read_cnt == 0)
		{
			return(0);
		}
		read_ptr = read_buf;
	}
	read_cnt --;
	*ptr = *read_ptr++;
	return(1);
}

ssize_t readn(int fd, void *vptr, size_t n)
{
	size_t		nleft;
	ssize_t		nread;
	char *ptr;

	ptr = (char *)vptr;
	nleft = n;
	while(nleft > 0)
	{
		if((nread = read(fd, ptr, nleft)) < 0)
		{
			if(errno == EINTR)
				nread = 0; // and call read() again.
			else
			{
				return (-1);
			}
		}else if(nread == 0)
			break;		//end of file
		nleft -= nread;
		ptr += nread;
	}
	return(n - nleft);		// return >= 0
}

ssize_t writen(int fd, const void *vptr, size_t n)
{
	size_t		nleft;
	ssize_t		nwritten;
	const char *ptr;

	ptr = (char *)vptr;

	nleft = n;
	while(nleft > 0)
	{
		if((nwritten = write(fd, ptr, nleft)) <= 0)
		{
			if(errno == EINTR)
				nwritten = 0;
			else
			{
				return (-1);
			}
		}
		nleft -= nwritten;
		ptr += nwritten;
	}
	return(n);
}

ssize_t ReadLine(register int fd, register char *vptr, register int maxlen)
{
	int		n, rc;
	char	c, *ptr;

	ptr = vptr;
	for(n = 1; n < maxlen; n ++)
	{
		if((rc = myread(fd, &c)) == 1)
		{
			*ptr++ = c;
			if(c == '\n') break; //newline is stored, just like fgets()
		}
		else if(rc == 0) //no more data read
		{
			if(n == 1) //EOF, no data read
				return(0);
			else break; //EOF, some data read
		}
		else
			return (-1); //error, error set by read()
	}
	*ptr = 0; //null terminate like fgets();
	if(c == '\n')
		return (n);
	else
		return (n - 1);	//because n = # bytes read + 1
}


ssize_t MyRecvFrom(int s, void *buf, size_t len, int flags, struct sockaddr *from, socklen_t *fromlen)
{
	ssize_t recvBytes = recvfrom(s, buf, len, flags, from, fromlen);
	if(recvBytes == 0 || recvBytes == -1)
	{
		printf("\nrecvfrom error: ");
		if(errno == EBADF)
			printf("The argument sockfd is an invalid descriptor.\n\n");
		else if(errno == ECONNREFUSED)
			printf("Remote host refused to allow connection.\n\n");
		else if(errno == ENOTSOCK)
			printf("The argument sockfd does not refer to a socket.\n\n");
		else if(errno == EAGAIN)
			printf("Marked non-blocking but operation would block.\n\n");
		else if(errno == EINTR)
			printf("Receive was interrupted.\n\n");
		else if(errno == EFAULT)
			printf("Receive buffer pointer points outside the process addr.\n\n");
		else if(errno == EINVAL)
			printf("Invalid argument passed.\n\n");
		else if(errno == ENOMEM)
			printf("Could not allocate memory.\n\n");
		else
			printf("Other causes.\n\n");

		fflush(stdout);
		close(s);
		exit(0);
	}
	return recvBytes;
}


UINT64 get_usec_since(struct timeval *old_time)
{
    struct timeval now;
    UINT64      result = 0;

    /* get the current time */
    gettimeofday(&now, NULL);

    /* return the elapsed time */
    while (now.tv_sec > old_time->tv_sec) {
	result += 1000000;
	--now.tv_sec;
    }
    return result + (now.tv_usec - old_time->tv_usec);

    /*------------------------------------------------------------
     * We used to calculate it like this, but the above is
     * usually a bit faster in the general case.  Note that we
     * have BIG problems if old_time is in the future, however...
     *------------------------------------------------------------*/
    /* return 1000000LL * (now.tv_sec - old_time->tv_sec) + now.tv_usec - old_time->tv_usec; */
}


/*------------------------------------------------------------------------
 * void usleep_that_works(u_int64_t usec);
 *
 * Sleeps for the given amount of microseconds, with better accuracy
 * than that offered by the standard usleep() routine.  We do a real
 * sleep until we get into within 10,000 microseconds, then busy-wait
 * for the rest of the time.
 *------------------------------------------------------------------------*/
void usleep_that_works(UINT64 usec)
{
    UINT64      sleep_time = (usec / 10000) * 10000;  /* the amount of time to sleep */
    struct timeval delay, now;

    /* get the current time */
    gettimeofday(&now, NULL);

    /* do the basic sleep */
    delay.tv_sec  = sleep_time / 1000000;
    delay.tv_usec = sleep_time % 1000000;
    select(0, NULL, NULL, NULL, &delay);

    /* and spin for the rest of the time */
    while (get_usec_since(&now) < usec);
}
